#include<iomanip>
#include<sstream>

#include "PGBasics.hh"

#include "definitions.hh"
#include "Ps.hh"
#include "intersection.hh"
#include "car.hh"
#include "scenario.hh"
#include "sts.hh"


#include "FixedController.hh"
#include "FixedPhaseController.hh"
#include "FixedRPController.hh"
#include "BasicController.hh"
#include "BinaryController.hh"
#include "FactoredController.hh"
#include "ScatsController.hh"
#include "NeuralNet.hh"
#include "NeuralNetBatch.hh"
#include "MixedController.hh"
#include "CyclicPolicyBias.hh"
#include "NACTransform.hh"
#include "OLPomdp.hh"
#include "GPomdp.hh"
#include "LineSearchAlg.hh"
#include "STSOptim.hh"
#include "CRFController.hh"

using namespace libpg;

#define CONTROLLER_OUTPUT_DIM 1
#define CRFCONTROLLER_OUTPUT_DIM 4
#define RANDOM_PHASE_CONTROLLER_OUTPUT_DIM 4

void make_scenario6_controllers(FactoredController::Controllers& controllers, 
				Vector& nnDims, Vector& squash);

Controller * make_crf_controller(int argc, char ** argv, STS *simulator);


STSOptim::STSOptim(int argc, char ** argv) 
{
	cout << "Process ID: " << getpid() << endl;
	srandom(Ps::pSEED);
	stepsPerEpoch = Ps::pSTEPS_PER_EPOCH;

	exit_check_window = Ps::pEXIT_CHECK_WINDOW;
	record_value = new double [exit_check_window];

	// This new constructor will build a default grid base map.
	simulator = new STS(new Scenario(Ps::pROWS, Ps::pCOLS));
    
	Vector nnDims;
	Vector squash;

	if (!Ps::pHIDDEN) 
	{
		nnDims.resize(LINEAR_NETWORK_LAYERS);
		nnDims[0] = simulator->getObsRows();
		nnDims[1] = CONTROLLER_OUTPUT_DIM;
		squash.resize(LINEAR_NETWORK_LAYERS); 
	} 
	else 
	{
		nnDims.resize(MLP_NETWORK_LAYERS);
		nnDims[0] = simulator->getObsRows();
		nnDims[1] = Ps::pHIDDEN;
		nnDims[2] = CONTROLLER_OUTPUT_DIM;
		squash.resize(MLP_NETWORK_LAYERS); 
		squash[1] = 1.0; 
	}
	
    // [daa] Why was the below put in?
    // squash[1] = false;
    

	FactoredController::Controllers controllers;

	switch(Ps::pCONTROLLER) 
	{
	case STS::_LearningController: //0

		// Build a learning controller for each intersection.	
		for (int i=0; i < (int)((STS*)simulator)->intersections.size(); i++) 
		{
			if(!((STS*)simulator)->intersections.at(i)->is_dummy()) 
			{
				if (!Ps::pNAC) 
					controllers.push_back(new BinaryController(new NeuralNet(nnDims, squash)));
				else 
					controllers.push_back(new NACTransform(new BinaryController(new NeuralNetBatch(nnDims, squash)), Ps::pNAC_DISCOUNT));
			}
			else 
				controllers.push_back(new FixedController(OBSERVATIONS_SCATS, 1,1,1,1));
		}
		break;

	case STS::_FixedPhaseController: //1
    
		// This controller gives the same fixed time to each phase
		for (int i=0; i < (int)((STS*)simulator)->intersections.size(); i++) 		
			controllers.push_back(new FixedPhaseController(OBSERVATIONS_SCATS, Ps::pFIXED_PHASE_LENGTH));
		break;

	case STS::_ScatsController: //2

		// This controller is modelled after the real SCATS controllers
		for (int i=0; i < (int)((STS*)simulator)->intersections.size(); i++) 
		{
			if(!((STS*)simulator)->intersections.at(i)->is_dummy()) 
			{
				controllers.push_back(new ScatsController
			      (OBSERVATIONS_SCATS,
			       Ps::pFIXED_PHASE_LENGTH,
			       Ps::pUPPER_THRESHOLD,
			       Ps::pLOWER_THRESHOLD,
			       Ps::pBOOST_SCATS));
			}
			else
			{
				controllers.push_back(new FixedController
				  (OBSERVATIONS_SCATS, 
				   1,1,1,1));
			}
		}
		break;

	case STS::_FixedController: //3

		// This controller gives a fixed time to each phase
		for (int i=0; i < (int)((STS*)simulator)->intersections.size(); i++) 
		{
			controllers.push_back(new FixedController
					  (OBSERVATIONS_SCATS, 
					   7,1,7,1));
		}
		break;

	case STS::_MixedController: //4
		// in this case, nnDims[0] is set wrongly to obsRows, since obsRows is
		// observation for learner + observation for SCATS: here, we correct:
 
		nnDims[0] -= OBSERVATIONS_SCATS;
		for (int i=0; i < (int)((STS*)simulator)->intersections.size(); i++) 
		{
			if(!((STS*)simulator)->intersections.at(i)->is_dummy()) 
			{
				NeuralNet* nn = new NeuralNet(nnDims, squash);
				controllers.push_back(new MixedController
				  (nn,
				   new BinaryController(nn),
				   new ScatsController
				   (OBSERVATIONS_SCATS,
				    Ps::pFIXED_PHASE_LENGTH,
				    Ps::pUPPER_THRESHOLD,
				    Ps::pLOWER_THRESHOLD,
				    Ps::pBOOST_SCATS, false),
				   Ps::pPRE_LEARNING_TIME));
			}
			else 
			{
				controllers.push_back(new FixedController
				  (OBSERVATIONS_SCATS, 
				   1,1,1,1));
			}
		}
		break;
    case STS::_RandomPhaseController: //5
		// in this case, the controller output dimension is not set correctly for the NN,
		// correct this:
		if (!Ps::pHIDDEN) 
		{
			nnDims[1] = RANDOM_PHASE_CONTROLLER_OUTPUT_DIM;
		} 
		else 
		{
			nnDims[2] = RANDOM_PHASE_CONTROLLER_OUTPUT_DIM;
		}
		
		// Build a learning controller for each intersection.	
		for (int i=0; i < (int)((STS*)simulator)->intersections.size(); i++) 
		{
			if(!((STS*)simulator)->intersections.at(i)->is_dummy()) 
			{
				if (!Ps::pNAC) controllers.push_back(new BasicController 
						 (new CyclicPolicyBias(new NeuralNet(nnDims, squash), 
								       Ps::pRPC_CYCLE_LENGTH, 
								       Ps::pBIAS)));
				else 
					controllers.push_back(new NACTransform(new BasicController (new CyclicPolicyBias(new NeuralNetBatch(nnDims, squash),Ps::pRPC_CYCLE_LENGTH, Ps::pBIAS)), Ps::pNAC_DISCOUNT));					 
			}
			else 
			{
				controllers.push_back(new FixedController
				    (OBSERVATIONS_SCATS, 
				     1,1,1,1));
			}
		}
		break;
      
    case STS::_Scenario6Controller: //6
		make_scenario6_controllers(controllers, nnDims, squash);
		break;

    case STS::_Scenario4Controller: // 7, mixed RP controller with fixed instead of SCATS trainer
		// in this case, nnDims[0] is set wrongly to obsRows, since obsRows is
		// observation for learner + observation for SCATS: here, we correct:
      
		nnDims[0] -= OBSERVATIONS_SCATS;
		// in this case, the controller output dimension is not set correctly for the NN,
		// correct this:
		
		if (!Ps::pHIDDEN) 
		{
			nnDims[1] = RANDOM_PHASE_CONTROLLER_OUTPUT_DIM;
		}
		else 
		{
			nnDims[2] = RANDOM_PHASE_CONTROLLER_OUTPUT_DIM;
		}
		
		for (int i=0; i < (int)((STS*)simulator)->intersections.size(); i++) 
		{
			std::vector<int>* opt_phases2 = new std::vector<int>();
			opt_phases2->push_back(1);
			opt_phases2->push_back(3);
			opt_phases2->push_back(2);
			opt_phases2->push_back(2);
			opt_phases2->push_back(2);
			opt_phases2->push_back(2);
			opt_phases2->push_back(2);
			opt_phases2->push_back(0);
			
			if(!((STS*)simulator)->intersections.at(i)->is_dummy()) 
			{
				NeuralNet* nn = new NeuralNet(nnDims, squash);
				controllers.push_back(new MixedController
					(nn, new BasicController(nn),
					new FixedRPController
					(OBSERVATIONS_SCATS, *opt_phases2),
					Ps::pPRE_LEARNING_TIME));
			}
			else 
			{
				controllers.push_back(new FixedController
					(OBSERVATIONS_SCATS, 
					1,1,1,1));
			}
		}
		break;
      
    case STS::_MixedRPController: //8 especially for BIG GRID scenario!
		// in this case, nnDims[0] is set wrongly to obsRows, since obsRows is
		// observation for learner + observation for SCATS: here, we correct:    
		
		nnDims[0] -= OBSERVATIONS_SCATS;
		// in this case, the controller output dimension is not set correctly for the NN,
		// correct this:
		
		if (!Ps::pHIDDEN) 
		{
			nnDims[1] = RANDOM_PHASE_CONTROLLER_OUTPUT_DIM;
		} 
		else 
		{
			nnDims[2] = RANDOM_PHASE_CONTROLLER_OUTPUT_DIM;
		}

		for (int i=0; i < (int)((STS*)simulator)->intersections.size(); i++) 
		{
			std::vector<int>* opt_phases = new std::vector<int>();
			opt_phases->push_back(0);
			opt_phases->push_back(0);
			opt_phases->push_back(0);
			opt_phases->push_back(0);
			opt_phases->push_back(1);
			opt_phases->push_back(1);
			opt_phases->push_back(1);
			opt_phases->push_back(1);
			opt_phases->push_back(2);
			opt_phases->push_back(2);	
			opt_phases->push_back(2);
			opt_phases->push_back(2);	
			opt_phases->push_back(3);
			opt_phases->push_back(3);
			opt_phases->push_back(3);
			opt_phases->push_back(3);
			if(!((STS*)simulator)->intersections.at(i)->is_dummy()) 
			{
				NeuralNet* nn = new NeuralNet(nnDims, squash);
				controllers.push_back(new MixedController
					(nn, new BasicController(nn),
						new FixedRPController(OBSERVATIONS_SCATS, *opt_phases),
						Ps::pPRE_LEARNING_TIME));
	  
				  /*	   new ScatsController
					(OBSERVATIONS_SCATS,
					Ps::pFIXED_PHASE_LENGTH,
					Ps::pUPPER_THRESHOLD,
					Ps::pLOWER_THRESHOLD,
					Ps::pBOOST_SCATS, true),
					Ps::pPRE_LEARNING_TIME)); */
			}
			else 
			{
				controllers.push_back(new FixedController
					(OBSERVATIONS_SCATS, 
					1,1,1,1));
			}
		}
		break;

    case STS::_CRFController:	    // 9
		controller = make_crf_controller(argc, argv, static_cast<STS *>(simulator));
		break;
    default:
		cout<<"!! Unknown controller type "<<Ps::pCONTROLLER;
		exit(EXIT_FAILURE);	
    }

    // The true in FactoredController tells it to split observations by column
	if (Ps::pCONTROLLER != STS::_CRFController) 
		controller = new FactoredController(controllers, true, Ps::pLOCAL_REWARDS);

	controller->setDiscount(Ps::pDISCOUNT); 
	controller->setStepSize(Ps::pSTEP_SIZE);
	epochsBetweenStepUpdates=0; // Turn off auto-step adaptation
	if(Ps::pBASELINE) 
	{
		useAutoBaseline = true;
	} 
	else 
	{
		useAutoBaseline = false;
	}
	baseline.resize(simulator->getRewardDim());

	// Initialize the observation class.
	obs.init(simulator->getObsRows(),simulator->getObsCols(),simulator->getAgents());

	// Reset everything
	resetLearning(); 

	if (Ps::pMAX_RAND > 0.0) 
		controller->randomizeParams(Ps::pMAX_RAND);
	else if (Ps::pHIDDEN) 
	{
		cout<<"!! Multi-layer perceptrons need MAX_RAND > 0\n";
		exit(EXIT_FAILURE);
	}

	// Save parameters when we achieve a new best.
	saveBest(Ps::pBEST_PARAMS_FILE);
    
	if (Ps::pRESTORE) read(Ps::pRESTORE_FILE);

#ifdef LINESEARCH
	stepSize = Ps::pSTEP_SIZE;
	lineSearchInit(Ps::pLINE_SEARCH_STEPS,
					Ps::pTOLERANCE,
					Ps::pBACK_OFF,
					Ps::pEXPONENT_BASE,
					Ps::pDOWNHILL_THRESH,
					Ps::pMAX_TR IES);
#endif
}



void STSOptim::printPerformanceInfo(bool printTitles, int steps, 
				    double avgReward, 
				    double maxParam, int seconds) 
{
	STS* sim = (STS*) simulator;

	std::vector<string>& statisticsNames = sim->getStatisticsNames();

	//Vector statistics(statisticsNames.size());
	Vector& statistics = sim->getStatistics(stepsPerEpoch);

	if(printTitles) 
	{
		cout<<"       "
			<<setw(CWIDTH)<<"Steps"
			<<setw(CWIDTH)<<"R"
			<<setw(CWIDTH)<<"Seconds";
		for(int i = 0; i < (signed int) statisticsNames.size(); i++) 
		{
			cout <<setw(CWIDTH)<< statisticsNames[i];
		}
		cout<<setw(CWIDTH)<<"max(Param)";
		cout << endl;
	}

	cout<<"Trial: "
		<<setw(CWIDTH)<<steps
		<<setw(CWIDTH)<<avgReward
		<<setw(CWIDTH)<<seconds;

	for(int i = 0; i < (signed int) statisticsNames.size(); i++) 
		cout <<setw(CWIDTH) <<statistics[i];

	cout<<setw(CWIDTH)<<maxParam;

	char **argv;
	int argc;
	
	argv = new char *[3];		// maximum 255 arguments to the sampler initializer
	argc = 1;

	argv[0] = new char[1];
	argv[0][0] = '\0';
	Utility::add_arg(&argc, argv, Ps::pSTEPS_PER_EPOCH, "-spe");		
	cout<<"   "<<controller->get_diagnosis(argc, argv)
		<<"   "<<simulator->get_diagnosis();
		
	Utility::destroy_all_args(argc, argv);		




	//////////////////////////////////////////////////////////////////////////
	// check whether we can exit now

	int cur_epoch = (steps + 2) / stepsPerEpoch;
	record_value[cur_epoch % exit_check_window] = statistics[4];
	if (exit_check_window > 0 && cur_epoch >= exit_check_window)
	{
		double max_rec, min_rec;
		max_rec = NINFINITY;
		min_rec = -NINFINITY;

		for (int i = 0; i < exit_check_window; i ++)
		{
			if (record_value[i] > max_rec)
				max_rec = record_value[i];
			if (record_value[i] < min_rec)
				min_rec = record_value[i];
		}

		if (max_rec - min_rec < 0.1 && fabs((max_rec - min_rec) / min_rec) < Ps::pEXIT_TOLERANCE)
			exit_flag = true;
		cout<<" :: "<<max_rec - min_rec<<" : "<<fabs((max_rec - min_rec) / min_rec)<<endl;
	}
	else
		cout << endl;

	//////////////////////////////////////////////////////////////////////////



	sim->clearStatistics();
	//sim->reset();
  
	if(steps == 0) 
	{
		// cout << "reseting simulator" << endl;
		sim->reset();

		// for special cases: save policy for intersection 12 during some time window
		if(Ps::pSAVE_FROM > -2) 
		{
			//if((signed int) sim->get_intersections().size() > Ps::pSAVE_INTERSECTION_POLICY) {
			//if((signed int) sim->get_intersections().size() > 6) {
			int save_from;

			string policy_suffix = "";
			if(Ps::pSAVE_FROM > -1) 
			{
				save_from = Ps::pSAVE_FROM;
			}
			else if(Ps::pCONTROLLER == STS::_ScatsController) 
			{
				save_from = 20000;
				//policy_suffix = "_scats";
			}
			else 
			{
				save_from = 10000;
				//save_from = 50000;
				//policy_suffix = "_nn";
			}

			int save_to = save_from + 1000;
			string filename = "policy"+ policy_suffix;
			IntersectionVec tracked;
			std::vector<int> want_to_track;

			//if(Ps::pSCENARIO == Scenario::BIG_GRID) {
			//want_to_track.push_back(13);
			//save_from = 10;
			//}
			if(Ps::pSCENARIO == Scenario::SINUSOIDAL || Ps::pSCENARIO == Scenario::SUDDEN_INFLUX) 
			{
				want_to_track.push_back(1);
				want_to_track.push_back(3);
				want_to_track.push_back(4); // cross, main intersec for 5x5
				want_to_track.push_back(5);
				want_to_track.push_back(7);

				//want_to_track.push_back(2); // main intersection for 3x3
				//save_from = 1;
			}
			if(Ps::pSCENARIO == Scenario::ONE_ROW_OFFSET_TEST) 
			{
				want_to_track.push_back(1);
				want_to_track.push_back(2);
				want_to_track.push_back(3);
			}
			if(Ps::pSCENARIO == Scenario::MAIN_STREET)	// scenario 6
			{ 
				
				want_to_track.push_back(2);
				want_to_track.push_back(3);
				want_to_track.push_back(5);
				want_to_track.push_back(6);
			}
			for(int i = 0; i < (signed int)want_to_track.size(); i++) 
			{
				int t = want_to_track.at(i);
				if((signed int)sim->get_intersections().size() > t) 
				{
					tracked.push_back(sim->get_intersection(t));
				}
				else
					cout << "cannot track intersection "<< t << ", does not exist!"<< endl;
			}

			//tracked.push_back(sim->get_intersection
			//(Ps::pSAVE_INTERSECTION_POLICY));

			sim->track_intersections(tracked);
			sim->save_history(save_from, save_to, filename);
			//}
		}
	}
 
	char fname[100];
	sprintf(fname, "%s.%d", Ps::pPARAM_FILE, getpid());
	write(fname);
/*
  if (cur_step > 200000)
	{
	  write(saveName);
		maxNodeRand = -1;
	}
*/  //write(Ps::pPARAM_FILE);
}


void make_scenario6_controllers(FactoredController::Controllers& controllers, 
					  Vector& nnDims, Vector& squash) {
  // NOTE: to speed up learning (since scenario 6 requires global rewards), some intersections
  // are controlled by FixedControllers (the end points of streets).
  // Scats and NN cannot be mixed, because observations are homogeneous across controllers in a scenario.
  // If Scats is to be used instead of NN below, getObservation() in sts.cc must be changed to return 
  // scats observations.

  /*
  for(int i = 2; i < 3; i++) {
    controllers.push_back(new ScatsController  // learns 5,2,8,1 or 4,1,10,1 (bad)
			  (OBSERVATIONS_SCATS,
			   Ps::pFIXED_PHASE_LENGTH,
			   Ps::pUPPER_THRESHOLD,
			   Ps::pLOWER_THRESHOLD,
			   Ps::pBOOST_SCATS));
  }
  */
  // performance with RL 3 and 1.2 cars per time step in each direction: 
  // travel time 14.7 (after restore),
  // throughput 1.35, cars 88.6, blocked 0
  
  controllers.push_back(new FixedController  // 20
  		(OBSERVATIONS_SCATS, 
  			 1,1,1,1));

 controllers.push_back(new FixedController  // 01
  		(OBSERVATIONS_SCATS, 
  			 1,1,1,1));

  controllers.push_back(new FixedController  // 11
  		(OBSERVATIONS_SCATS, 
		 1,3,5,1)); //Scats learns 1,1,1,1!! //2,4?

  
  controllers.push_back(new FixedController  // 21
			(OBSERVATIONS_SCATS, 
  			 4,2,6,1));  
 
  controllers.push_back(new FixedController  // 31
  		(OBSERVATIONS_SCATS, 
  			 1,1,1,1));

  controllers.push_back(new FixedController  // 12
  		(OBSERVATIONS_SCATS, 
		 12,2,1,1)); //12,2,1,1

  controllers.push_back(new FixedController  // 22
  		(OBSERVATIONS_SCATS, 
		 6,1,2,1)); 

  for(int i = 7; i < 9; i++) { // 13 and 23
  controllers.push_back(new FixedController
  		(OBSERVATIONS_SCATS, 
  			 1,1,1,1));
  }
}


Controller * make_crf_controller(int _argc, char ** _argv, STS *simulator)
{
	Controller *controller;
	int argc, num_edge_fea, matsq;
	char ** argv;

	argv = new char * [255];		// maximum 255 arguments to the sampler initializer
	Utility::copy_arg(&argc, argv, _argc, _argv);

	Utility::add_arg(&argc, argv, Ps::pDISCOUNT, "-dc");
	Utility::add_arg(&argc, argv, Ps::pSTEP_SIZE, "-ss1");	
	Utility::add_arg(&argc, argv, Ps::pNAC_DISCOUNT, "-td");	
	Utility::add_arg(&argc, argv, Ps::pSTEPS_PER_EPOCH, "-spe");	
	Utility::add_arg(&argc, argv, Ps::pMAX_STEPS, "-xs");	
	Utility::add_arg(&argc, argv, Ps::pMAX_TIME, "-xt");	
	Utility::add_arg(&argc, argv, Ps::pEDGE_FEA_SEL, "-efs");
	Utility::add_arg(&argc, argv, Ps::pEDGE_REWARD_DEF, "-erd");

	parse_para(&argc, argv);

	if (Ps::pSTEP_EDGE >= 0)
		Utility::add_arg(&argc, argv, Ps::pSTEP_EDGE, "-ss2");
	if (Ps::pSTEP_ENODE >= 0)
		Utility::add_arg(&argc, argv, Ps::pSTEP_ENODE, "-ss3");

	num_edge_fea = 0;

	if (Ps::pCONTROLLER == STS::_CRFController)
	{
		Utility::add_arg(&argc, argv, 4, "-app");

		Utility::add_arg(&argc, argv, Ps::pMIN_DELAY, "-idy");
		Utility::add_arg(&argc, argv, Ps::pMAX_DELAY, "-xdy");

		matsq = CRFCONTROLLER_OUTPUT_DIM * CRFCONTROLLER_OUTPUT_DIM;
		if (Ps::pEDGE_FEA_SEL & 1)
		{
			cout<<"Using binary constant edge bias = true\n";
			num_edge_fea += 1;
		}
		else
			cout<<"Using binary constant edge bias = false\n";

		if (Ps::pEDGE_FEA_SEL & 2)
		{
			cout<<"Using square matrix constant edge bias = true\n";
			num_edge_fea += matsq;
		}
		else
			cout<<"Using square matrix constant edge bias = false\n";

		if (Ps::pEDGE_FEA_SEL & 4)
		{
			cout<<"Using binary (past, current) = true\n";
			num_edge_fea += Ps::pMAX_DELAY - Ps::pMIN_DELAY + 1;
		}
		else
			cout<<"Using binary (past, current) = false\n";

		if (Ps::pEDGE_FEA_SEL & 8)
		{
			cout<<"Using square matrix (past, current) = true\n";
			num_edge_fea += matsq * (Ps::pMAX_DELAY - Ps::pMIN_DELAY + 1);
		}
		else
			cout<<"Using square matrix (past, current) = false\n";

		if (Ps::pEDGE_FEA_SEL & 16)
		{
			cout<<"Using binary (current, past) = true\n";
			num_edge_fea += Ps::pMAX_DELAY - Ps::pMIN_DELAY + 1;
		}
		else
			cout<<"Using binary (current, past) = false\n";

		if (Ps::pEDGE_FEA_SEL & 32)
		{
			cout<<"Using square matrix (current, past) = true\n";
			num_edge_fea += matsq * (Ps::pMAX_DELAY - Ps::pMIN_DELAY + 1);
		}
		else
			cout<<"Using square matrix (current, past) = false\n";
	}
	Utility::add_arg(&argc, argv, num_edge_fea, "-nef");

	switch(Ps::pSCENARIO)
	{
		case Scenario::ONE_ROW_OFFSET_TEST:
			Utility::add_arg(&argc, argv, 1, "-tid");		// chain
			break;
/*        
		case Scenario::SUDDEN_INFLUX:
		case Scenario::SINUSOIDAL:
			Utility::add_arg(&argc, argv, 4, "-tid");		// cross
			Utility::add_arg(&argc, argv, 1 + simulator->get_cols() / 2, "-chi");
			Utility::add_arg(&argc, argv, simulator->get_rows() / 2, "-ncu");
			break;

		case Scenario::BIG_GRID:
			Utility::add_arg(&argc, argv, 2, "-tid");		// grid
			break;

		case Scenario::EW_BIASED_CENTER_ONLY:
			cout<<"Controller for EW_BIASED_CENTER_ONLY hasn't been implemented yet.  Abort.\n";
			exit(1);
			break;

		case Scenario::MAIN_STREET:
			cout<<"Controller for MAIN_STREET hasn't been implemented yet.  Abort.\n";
			exit(1);
			break;
*/
		default:
			cout<<"In make_crf_controller, unforseen senario: "<<Ps::pSCENARIO<<".  Abort.\n";
			exit(1);
			break;
	}

	if (Ps::pMAX_TREE_NO >= 0)	
		Utility::add_arg(&argc, argv, Ps::pMAX_TREE_NO, "-xti");
	if (Ps::pMAX_TREE_TIME >= 0)
		Utility::add_arg(&argc, argv, Ps::pMAX_TREE_TIME, "-xtt");
	if (Ps::pMAX_RAND_EDGE >= 0)
		Utility::add_arg(&argc, argv, Ps::pMAX_RAND_EDGE, "-xe");
	Utility::add_arg(&argc, argv, Ps::pMAX_RAND, "-xn");

	Utility::add_arg(&argc, argv, simulator->get_rows(), "-nr");
	Utility::add_arg(&argc, argv, simulator->get_cols(), "-nc");
	Utility::add_arg(&argc, argv, simulator->getAgents(), "-na");
	Utility::add_arg(&argc, argv, CRFCONTROLLER_OUTPUT_DIM, "-ncl");
	Utility::add_arg(&argc, argv, simulator->getObsRows(), "-no");
	Utility::add_arg(&argc, argv, 0, "-nnxf");

	if (Ps::pLOCAL_REWARDS)
		Utility::add_arg(&argc, argv, 1, "-ulr");
	else
		Utility::add_arg(&argc, argv, 0, "-ulr");

	Utility::add_arg(&argc, argv, Ps::pRPC_CYCLE_LENGTH, "-cyt");
	Utility::add_arg(&argc, argv, Ps::pBIAS, "-bias");

	controller = new CRFController(&argc, argv);
	if (Ps::pNAC)
		controller = new NACTransform(controller, Ps::pNAC_DISCOUNT);
//	else
//		Ps::pBASELINE = false;

	return controller;
}

