
// $Id$

#include<PGBasics.hh>
#include<TreeBasics.hh>

#include"SensorSim.hh"
#include<iomanip>

// Defaults
#define OBS_DIM 2
#define REWARD_FOCUS -1
#define REWARD_KILL 30
#define NUM_CELL 3
#define NUM_TARGET 2 

#define DBG_POL false

using namespace std;
using namespace libpg;


/**
 * @param SensorSize how many paths in for each direction, e.g., square
 * root of number of agents.
 * @param frequency with which packets arrive. 
 * @param bufferSize new packets will be dropped if more than this
 * packets waiting in any queue
 */
SensorSim::SensorSim(int *argc, char **argv) 
{
	int app_id;

	Utility::get_arg(&num_cell, argc, argv, "-nc", 1, NUM_CELL);
	Utility::get_arg(&num_target, argc, argv, "-nt", 1, NUM_TARGET);
	Utility::get_arg(&ctl_id, argc, argv, "-ctl", 1, 0);			// 1 for NN, 2 for CRF Chain, 3 for CRF circle
	Utility::get_arg(&max_cycle, argc, argv, "-xy", 0, 0);
	Utility::get_arg(&num_epoch_per_review, argc, argv, "-spe", 0, 50);
	Utility::get_arg(&app_id, argc, argv, "-app", 1, 0);
	Utility::add_arg(argc, argv, 2, "-no");

	num_sensor = 2 * num_cell + 2;
	total_killed = total_cycle = 0;
	total_cycle = 0;
	cur_epoch = 1;

	if (ctl_id == 2)
		ctl_id = app_id;


	if (ctl_id == 1)					// 1: neural network
	{
		num_agent = num_sensor;
	}
	else if (ctl_id == 2)				// 2: for CRF Chain
	{
		num_agent = num_cell + 1;
		Utility::add_arg(argc, argv, 1, "-nnxf");
		Utility::add_arg(argc, argv, 6, "-ncl");
		Utility::add_arg(argc, argv, 1, "-nef");
	}
	else if (ctl_id == 3)				// 3: CRF circle
	{
		num_agent = num_sensor;
		Utility::add_arg(argc, argv, 1, "-nnxf");
		Utility::add_arg(argc, argv, 3, "-ncl");
		Utility::add_arg(argc, argv, 3, "-nef");
	}

	Utility::add_arg(argc, argv, num_agent, "-na");

	occupied = new bool [num_cell + 2];
	energy = new int [num_target + 1];
	position = new int [num_target + 1];
	
    // randomly locate the position of targets
	buffer1 = new int [num_cell + 1];
	buffer2 = new int [num_cell + 1];

	InitializeGame();

	restartGame();

    cout<<"Initialized SensorSim. Num_Target = "<<num_target<<", num_cell = "<<num_cell<<endl;
	cout<<"max_cycle = "<<max_cycle<<", num_epoch_per_review = "<<num_epoch_per_review<<". ";
	if (ctl_id == 1)
		cout<<"Using NN.";
	else if (ctl_id == 2)
		cout<<"Using CRF Chain.";
	else if (ctl_id == 3)
		cout<<"Using CRF circle.";
	cout<<endl;
}


SensorSim::~SensorSim()
{
	delete [] occupied;
	delete [] energy;
	delete [] position;
	delete [] buffer1;
	delete [] buffer2;

	for (int l = 1; l <= num_epoch_per_review; l ++)
	{
		delete [] store_occupied[l];
		delete [] store_position[l];
	}
	delete [] store_position;
	delete [] store_occupied;
}

inline int SensorSim::getObsRows() { return OBS_DIM; }

inline int SensorSim::getObsCols() { return num_agent; }

inline int SensorSim::getAgents() { return num_agent; }

inline int SensorSim::getActionDim() { return num_agent; }

inline double SensorSim::get_max_reward() { return num_target * REWARD_KILL - 3 * 3 * num_target;  }

/**
 * Global rewards 
 */
inline int SensorSim::getRewardDim() { return 1; }

/**
 * Just return the next reward set in doAction
 */
void SensorSim::getReward(Vector& rewards)
{
    assert(rewards.size() == 1);
    rewards[0] = nextReward;
}


void SensorSim::InitializeGame()
{
	int i, j, k, l, minimum;

	store_occupied = new bool *[num_epoch_per_review + 1];
	store_position = new int *[num_epoch_per_review + 1];

	for (l = 1; l <= num_epoch_per_review; l ++)
	{
		store_occupied[l] = new bool [num_cell + 1];
		store_position[l] = new int [num_target + 1];

		for (i =1; i <= num_cell; i ++)
			occupied[i] = false;

		for (i = 1; i <= num_cell; i ++)	
			buffer1[i] = buffer2[i] = rand();

		// then sort the random numbers in buffer2
		for (i = 1; i < num_cell; i ++)
		{
			minimum = buffer2[i];
			k = i;
			for (j = i + 1; j <= num_cell; j ++)
			{
				if (buffer2[j] < minimum)
				{
					k = j;
					minimum = buffer2[j];
				}
			}
			if (k != i)
			{
				j = buffer2[i];
				buffer2[i] = minimum;
				buffer2[k] = j;
			}
		}

		// Assign occupied[]
		for (i = 1; i <= num_target; i ++)
		{
			for (j = 1; j <= num_cell; j ++)
			{
				if (buffer1[i] == buffer2[j] && !occupied[j])
					break;
			}
			occupied[j] = true;
		}
		
		j = 1;
		for (i = 1; i <= num_target; i ++)
		{
			while(!occupied[j])
				j ++;
			position[i] = j ++;
		}

		for (i = 1; i <= num_cell; i ++)
			store_occupied[l][i] = occupied[i];

		for (i =1; i <= num_target; i ++)
			store_position[l][i] = position[i];			
	}
/*	for (i = 1; i <= num_target; i ++)
		printf("%d: %d\t", i, position[i]);
	printf("\n");
	for (i = 1; i <= num_cell; i ++)
		printf("%d: %d\t", i, occupied[i]);
	printf("\n");
*/
}


void SensorSim::restartGame()
{
	int i;

	num_living_target = num_target;

	for (i = 1; i <= num_cell; i ++)
		occupied[i] = store_occupied[cur_epoch][i];

	for (i =1; i <= num_target; i ++)
	{
		position[i] = store_position[cur_epoch][i];
		energy[i] = 3;
	}

	cur_cycle = 1;
}


/**
 * Two observations per node giving normalized number of waiting packets. 
 * @param obs observation structure that should be correct size already
 */
void SensorSim::getObservation(Observation& obs) 
{
	int cur_sensor, cur_cell;

    assert((int)obs.getFeatures().size1() == OBS_DIM);
	assert((int)obs.getFeatures().size2() == num_agent);
 	occupied[0] = false;
	occupied[num_cell + 1] = false;

	for (cur_sensor = 0; cur_sensor <= num_cell; cur_sensor ++)
	{
		obs.getFeatures()(0, cur_sensor) = (occupied[cur_sensor] ? 1 : 0);
		obs.getFeatures()(1, cur_sensor) = (occupied[cur_sensor + 1] ? 1 : 0);
	}

	if (ctl_id == 1 || ctl_id == 3)		// for distributed RL, and for CRF circle
	{
		for (cur_cell = 0; cur_sensor < num_sensor; cur_sensor ++, cur_cell ++)
		{
			obs.getFeatures()(0, cur_sensor) = (occupied[cur_cell] ? 1 : 0);
			obs.getFeatures()(1, cur_sensor) = (occupied[cur_cell + 1] ? 1 : 0);
		}	
	}
}


// which type of communication, 
// which determines format of observation to return, and action to take

/**
 * Do the suggested action, check targets being focused, calculate the rewards.
 * maybe restart the game if all targets are killed.  Otherwise, randomly move targets.
 */
int SensorSim::doAction(Vector& action) 
{
	int cur_target, cur_position, direction;

	occupied[0] = true;
	occupied[num_cell + 1] = true;
    nextReward = 0;

    assert((int)action.size() == num_agent);

	if (ctl_id == 1 || ctl_id == 3)		// NN or CRF circle of original
	{
		// Find the first living target.  Assuming there must exist a living target.
		// Otherwise, game restarted already.
		for (cur_target = 1; cur_target <= num_target; cur_target ++)
		{
			if (energy[cur_target])
			{
				cur_position = position[cur_target];

				if ( (action[cur_position - 1] == 2) + (action[cur_position] == 1) 
						+ (action[cur_position + num_cell] == 2) 
						+ (action[cur_position + num_cell + 1] == 1) >= 3)
				{
					energy[cur_target] --;
					if (!energy[cur_target])		// cur_target becomes dead
					{
						nextReward += REWARD_KILL;
						occupied[cur_position] = false;
						num_living_target --;
					}
				}
			}
		}
		for (cur_position = 0; cur_position < num_sensor; cur_position ++)
		{
			if (action[cur_position])
				nextReward += REWARD_FOCUS;
		}
	}
	else if (ctl_id == 2)		// for CRF supernode chain style controllers
	{
		for (cur_target = 1; cur_target <= num_target; cur_target ++)
		{
			if (energy[cur_target])
			{
				cur_position = position[cur_target];
				direction = 0;
				if (action[cur_position - 1] == 5)
				{
					if (action[cur_position] == 1 || action[cur_position] == 3 || action[cur_position] == 4)
						direction = 1;
				}
				else if (action[cur_position] == 4)
				{
					if (action[cur_position - 1] == 2 || action[cur_position - 1] == 3)
						direction = 1;
				}
				if (direction)		// being targeted.
				{
					energy[cur_target] --;
					if (!energy[cur_target])		// cur_target becomes dead
					{
						nextReward += REWARD_KILL;
						occupied[cur_position] = false;
						num_living_target --;
					}
				}
			}
		}		
		for (cur_position = 0; cur_position <= num_cell; cur_position ++)
		{
			if (action[cur_position] > 2)
				nextReward += 2 * REWARD_FOCUS;
			else if (action[cur_position])
				nextReward += REWARD_FOCUS;
		}
	}
	
	// No more targets alive
	if (num_living_target == 0 || max_cycle && cur_cycle >= max_cycle)
	{
		total_killed += num_target - num_living_target;
		total_cycle += cur_cycle;
		
		if (cur_epoch == num_epoch_per_review)
		{
			sprintf(diagnosis_msg, "%g : %g", 
					total_killed / (double) cur_epoch, total_cycle / (double) cur_epoch);
			total_killed = total_cycle = 0;			
			cur_epoch = 1;
		}
		else
			cur_epoch ++;

		restartGame();
		return 0;
	}

	// Now move targets
	cur_cycle ++;
	for (cur_target = 1; cur_target <= num_target; cur_target ++)
	{		
		if (energy[cur_target])		// cur_target is still alive
		{
			cur_position = position[cur_target];
			if (!occupied[cur_position - 1] || !occupied[cur_position + 1])
			{
				direction = rand() % 3;
				if (direction == 1 && !occupied[cur_position - 1])		// move left
				{
					position[cur_target] --;
					occupied[cur_position] = false;
					occupied[cur_position - 1] = true;
				}
				else if (direction == 2 && !occupied[cur_position + 1])		// move right
				{
					position[cur_target] ++;
					occupied[cur_position] = false;
					occupied[cur_position + 1] = true;
				}
			}
		}
	}

	if (DBG_POL) 
	{
		cout<<"Current Dead:\t";
		for (cur_target = 1; cur_target <= num_target; cur_target ++)
		{
			if (!energy[cur_target])
				cout<<cur_target<<"\t";
		}
		cout<<endl;

		cout<<"Current Position:\t";
		for (cur_target = 1; cur_target <= num_target; cur_target ++)
		{
			if (energy[cur_target])
				cout<<position[cur_target]<<"\t";
		}
		cout<<endl;
	}

    return -1;
}

