#ifndef LOGSUMEXP_hh
#define LOGSUMEXP_hh

/* exp2.h -- fast EXP, inline function versions
 *
 * (c) 1993-2000 Nici Schraudolph
 *     (all rights reserved)
 *
 * See: N. Schraudolph,
 * "A Fast, Compact Approximation of the Exponential Function",
 * Neural Computation 11:4 (1999), MIT Press,
 * ftp://ftp.idsia.ch/pub/nic/exp.ps.gz
 *
 * float:   sxxx|xxxx|xmmm|mmmm|mmmm|mmmm|mmmm|mmmm
 * ( 1.0 == 0011 1111 1000 0000 0000 0000 0000 0000 )
 *
 * works only with IEEE 754 floating point format!
 *
 * double:  sxxx|xxxx|xxxx|mmmm|mmmm|mmmm|mmmm|mmmm
 * ( Inf == s111 1111 1111 0000 0000 0000 0000 0000 ...)
 *          mmmm|mmmm|mmmm|mmmm|mmmm|mmmm|mmmm|mmmm
 *
 * works like a 2k entry lookup table with linear interpolation
 */

//#define i386

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "inttypes.h"

typedef union
{
    double d;
    int64_t l;
    struct
    {
#if defined(i386) || defined(__alpha)
/* #ifdef __LITTLE_ENDIAN */
        int32_t j, i;
#else
        int32_t i, j;
#endif
    } n;
    struct
    {
#if defined(i386) || defined(__alpha)
/* #ifdef __LITTLE_ENDIAN */
        uint32_t j, i;
#else
        uint32_t i, j;
#endif
    } u;
} _eco;

inline double fexp2(register double y)
{
    register _eco tmp;

    tmp.d = y;
    tmp.n.i += (20 << 20);
    tmp.n.i = (int32_t) tmp.d;
    tmp.n.i += (1023 << 20);
    tmp.n.j = 0;
    return tmp.d;
}

inline double fexp2l(register double y)
{
    register _eco tmp;

    tmp.d = y;
    tmp.n.i += (52 << 20);
    tmp.l = (int64_t) tmp.d;
    tmp.n.i += (1023 << 20);
    return tmp.d;
}

inline double flog2(register double y)
{
    register _eco tmp;

    tmp.d = y;
    tmp.n.i -= (1023 << 20);
    if (tmp.n.i)
    {
        tmp.d = (double) tmp.n.i;
        tmp.n.i -= (20 << 20);
    }
    return tmp.d;
}

inline double flog2l(register double y)
{
    register _eco tmp;

    tmp.d = y;
    tmp.n.i -= (1023 << 20);
    if (tmp.l)
    {
        tmp.d = (double) tmp.l;
        tmp.n.i -= (52 << 20);
    }
    return tmp.d;
}

/* EXP2(y) = EXP(y/2)/EXP(-y/2) */ 

inline double fexp2r(register double y)
{
    register _eco tmp, tmp2;

    tmp.d = y;
    tmp.n.i += (19 << 20);
    tmp2.n.i = (1023 << 20) - (int32_t) tmp.d;
    tmp.n.i = (1023 << 21) - tmp2.n.i;
    tmp.n.j = tmp2.n.j = 0;
    return tmp.d/tmp2.d;
}


inline double fexp2r_check(register double y)
{
    register _eco tmp, tmp2;

	if (y > 1010)		// largest acceptable exponent (base 2)
	{
		printf("Too large exponent: %f\n", y);
		exit(1);
	}
	else if (y < -1010)
		return 0.0;

    tmp.d = y;
    tmp.n.i += (19 << 20);
    tmp2.n.i = (1023 << 20) - (int32_t) tmp.d;
    tmp.n.i = (1023 << 21) - tmp2.n.i;
    tmp.n.j = tmp2.n.j = 0;
    return tmp.d/tmp2.d;
}


inline double fexp2rl(register double y)
{
    register _eco tmp, tmp2;

    tmp.d = y;
    tmp.n.i += (51 << 20);
    tmp.l = (int64_t) tmp.d;
    tmp2.l = -tmp.l;
    tmp.n.i += (1023 << 20);
    tmp2.n.i += (1023 << 20);
    return tmp.d/tmp2.d;
}

/* LOG2(y) = (LOG(y) - LOG(1/y))/2 */

inline double flog2r(register double y)
{
    register _eco tmp, tmp2;

    tmp.d = y;
    tmp2.d = 1.0/y;
    tmp.n.i -= tmp2.n.i;
    if (tmp.n.i)
    {
        tmp.d = (double) tmp.n.i;
        tmp.n.i -= (21 << 20);
    }
    return tmp.d;
}

inline double flog2rl(register double y)
{
    register _eco tmp, tmp2;

    tmp.d = y;
    tmp2.d = 1.0/y;
    tmp.l -= tmp2.l;
    if (tmp.l)
    {
        tmp.d = (double) tmp.l;
        tmp.n.i -= (21 << 20);
    }
    return tmp.d;
}



/*
int main(int argc, char **argv)
{
	if(argc > 1)	{
		printf("size of int: %d\n", sizeof(int));
		printf("size of float: %d\n", sizeof(float));
		printf("size of double: %d\n", sizeof(double));
	}
	
	printf("By library, exp(%lf) = %lf\n", atof(argv[1]), pow(2, atof(argv[1])));
	
	printf("By fexp2, exp(%lf) = %lf\n", atof(argv[1]), fexp2(atof(argv[1])));
	
	printf("By fexp2rl, fexp2rl(%lf) = %lf\n", atof(argv[1]), fexp2rl(atof(argv[1])));

	return 0;
}
*/
#endif
